
class ApiEndpoints(object):
    def __init__(self, version='v3'):
        self.version = version
        self.API_URL = 'https://api.helloworks.com'
        self.API_ENDPOINTS = {
            'v3': {
                'auth_token': 'token/{}',
                'create_workflow_instance': 'workflow_instances',
                'preview_workflow_instance': 'workflow_instances/preview',
                'cancel_workflow_instance': 'workflow_instances/{}/cancel',
                'get_workflow_instance': 'workflow_instances/{}',
                'get_workflow_instance_document': 'workflow_instances/{}/documents/{}',
                'get_workflow_instance_audit_trail': 'workflow_instances/{}/audit_trail',
                'get_workflow_instance_documents': 'workflow_instances/{}/documents',
                'get_workflow_instance_steps': 'workflow_instances/{}/steps',
                'get_document_link': 'workflow_instances/{}/document_link',
                'get_authenticated_link_for_workflow_instances_step': 'workflow_instances/{}/steps/{}/delegate_auth',
                'save_settings_with_logo_file': 'white_label/settings'
            }
        }

    def __generate_url(self, url):
        endpoint = f'{self.API_URL}/{self.version}/{self.API_ENDPOINTS[self.version][url]}'
        return endpoint

    def auth_token(self, api_key_id):
        return self.__generate_url('auth_token').format(api_key_id)

    def create_workflow_instance(self):
        return self.__generate_url('create_workflow_instance')

    def preview_workflow_instance(self):
        return self.__generate_url('preview_workflow_instance')

    def cancel_workflow_instance(self, workflow_instance_id):
        return self.__generate_url('cancel_workflow_instance').format(workflow_instance_id)

    def get_workflow_instance(self, workflow_instance_id):
        return self.__generate_url('get_workflow_instance').format(workflow_instance_id)

    def get_workflow_instance_document(self, workflow_instance_id, document_id):
        return self.__generate_url('get_workflow_instance_document').format(workflow_instance_id, document_id)

    def get_workflow_instance_audit_trail(self, workflow_instance_id):
        return self.__generate_url('get_workflow_instance_audit_trail').format(workflow_instance_id)

    def get_workflow_instance_documents(self, workflow_instance_id):
        return self.__generate_url('get_workflow_instance_documents').format(workflow_instance_id)

    def get_workflow_instance_steps(self, workflow_instance_id):
        return self.__generate_url('get_workflow_instance_steps').format(workflow_instance_id)

    def get_document_link(self, workflow_instance_id):
        return self.__generate_url('get_document_link').format(workflow_instance_id)

    def get_authenticated_link_for_workflow_instances_step(self, workflow_instance_id, step_id):
        return self.__generate_url('get_authenticated_link_for_workflow_instances_step').format(workflow_instance_id, step_id)

    def save_settings_with_logo_file(self):
        return self.__generate_url('save_settings_with_logo_file')



from .utils import ApiEndpoints, HwRequest, HwAccess
from .utils import _file_downloader, _clean_dict, _file_extension 
from .utils.exceptions import get_exception_message

class HwClient(object):

    def __init__(self, api_key_id, API_KEY_VALUE):
        self.request = HwRequest()
        self.endpoints = ApiEndpoints()
        self.access = HwAccess(api_key_id, API_KEY_VALUE)
        
    def create_workflow_instance(self,
                                 workflow_id,
                                 participants,
                                 merge_fields=None,
                                 callback_url=None,
                                 redirect_url=None,
                                 document_delivery=True,
                                 white_label_id=None,
                                 delegated_authentication=False,
                                 metadata=None,
                                 document_delivery_type=None):
        """
            Method to create an instance of a Workflow to send out to a user.
        """
        create_workflow_url = self.endpoints.create_workflow_instance()
        dict_response = self._post_workflow_instance(create_workflow_url,
                                            workflow_id=workflow_id,
                                            participants=participants,
                                            merge_fields=merge_fields,
                                            callback_url=callback_url,
                                            redirect_url=redirect_url,
                                            document_delivery=document_delivery,
                                            white_label_id=white_label_id,
                                            delegated_authentication=delegated_authentication,
                                            metadata=metadata,
                                            document_delivery_type=document_delivery_type)

        return dict_response


    def preview_workflow_instance(self,
                                  workflow_id,
                                  participants,
                                  merge_fields=None,
                                  callback_url=None,
                                  redirect_url=None,
                                  document_delivery=True,
                                  white_label_id=None,
                                  delegated_authentication=False,
                                  metadata=None,
                                  document_delivery_type=None):
        """
            Method to create a preview instance of a Workflow to send out to a user.
        """
        preview_workflow_url = self.endpoints.preview_workflow_instance()
        dict_response = self._post_workflow_instance(preview_workflow_url,
                                            workflow_id=workflow_id,
                                            participants=participants,
                                            merge_fields=merge_fields,
                                            callback_url=callback_url,
                                            redirect_url=redirect_url,
                                            document_delivery=document_delivery,
                                            white_label_id=white_label_id,
                                            delegated_authentication=delegated_authentication,
                                            metadata=metadata,
                                            document_delivery_type=document_delivery_type)
        return dict_response

    def cancel_workflow_instance(self, workflow_instance_id):
        """
            Method to cancel a Workflow Instance that is currently active
        """
        cancel_workflow_url = self.endpoints.cancel_workflow_instance(workflow_instance_id)
        token = self.access.get_jwt_token()
        response = self.request.put(cancel_workflow_url, token)
        if response.status_code != 200:
            exc_msg = get_exception_message(response)
            raise Exception(exc_msg)
        return response.json()

    def get_workflow_instance(self, workflow_instance_id):
        """
            Method to retrieve information on a single Workflow Instance
        """
        get_workflow_url = self.endpoints.get_workflow_instance(workflow_instance_id)
        token = self.access.get_jwt_token()
        response = self.request.get(get_workflow_url, token)
        if response.status_code != 200:
            exc_msg = get_exception_message(response)
            raise Exception(exc_msg)
        return response.json().get('data')

    def get_workflow_instance_document(self, workflow_instance_id, document_id, get_request=True):
        """
            Method to get a given document of a workflow instance
            If get_request is set to true it will return the request instead of writing the file
        """
        get_workflow_document_url = self.endpoints.get_workflow_instance_document(workflow_instance_id, document_id)
        return self._file_endpoints(get_workflow_document_url, f'{document_id}.pdf', get_request)


    def get_workflow_instance_audit_trail(self, workflow_instance_id, get_request=True):
        """
            Method to get the audit trail of a workflow instance
            If get_request is set to true it will return the request instead of writing the file
        """
        get_workflow_audit_trail_url = self.endpoints.get_workflow_instance_audit_trail(workflow_instance_id)
        return self._file_endpoints(get_workflow_audit_trail_url, f'{workflow_instance_id}-audit-trail.pdf', get_request)

    def get_workflow_instance_documents(self, workflow_instance_id, get_request=True):
        """
            Method to get the documents of a workflow instance
            If get_request is set to true it will return the request instead of writing the file
        """
        get_workflow_documents_url = self.endpoints.get_workflow_instance_documents(workflow_instance_id)
        return self._file_endpoints(get_workflow_documents_url, f'{workflow_instance_id}-files.zip', get_request)

    def get_workflow_instance_steps(self, workflow_instance_id):

        """
            Method to get the specified workflow instance, the step id, the role that will be completing the step,
            the signer's full name, and the unauthenticated url that can be used to start entering information.
        """

        get_workflow_instance_steps_url = self.endpoints.get_workflow_instance_steps(workflow_instance_id)
        token = self.access.get_jwt_token()
        response = self.request.get(get_workflow_instance_steps_url, token)
        if response.status_code != 200:
            exc_msg = get_exception_message(response)
            raise Exception(exc_msg)
        return response.json().get('data')

    def get_document_link(self, workflow_instance_id):
        """
            For the specified workflow instance, get the step id, the role that will be completing the step,
            the signer's full name, and the unauthenticated url that can be used to start entering information.
        """

        get_document_link = self.endpoints.get_document_link(workflow_instance_id)
        token = self.access.get_jwt_token()
        response = self.request.get(get_document_link, token)
        if response.status_code != 200:
            exc_msg = get_exception_message(response)
            raise Exception(exc_msg)
        return response.json().get('data')

    def get_authenticated_link_for_workflow_instance_step(self, workflow_instance_id, step_id):
        """
            Method to get the authenticated link for a workflow instance step
        """
        get_authentiated_link_for_workflow_instances_step_url = self.endpoints.get_authenticated_link_for_workflow_instances_step(workflow_instance_id, step_id)
        token = self.access.get_jwt_token()
        response = self.request.get(get_authentiated_link_for_workflow_instances_step_url, token)
        if response.status_code != 200:
            exc_msg = get_exception_message(response)
            raise Exception(exc_msg)
        return response.json().get('data')

    def save_setting_with_logo_file(self,
                                    logo_file,
                                    logo_name=None,
                                    white_label_id=None,
                                    primary_color=None,
                                    logo_hidden=True,
                                    team_name=None,
                                    workflow_name=None,
                                    stepready_email_text=None):
        """
            Method to create a white label ID to later be used in the workflow
            creation.
        """
        token = self.access.get_jwt_token()
        white_label_url = self.endpoints.save_settings_with_logo_file()
        extension = _file_extension(logo_file, logo_name)
        file = ('logo', logo_file, f'image/{extension}')
        dict_response = self._post_files(white_label_url,
                                         logo_file=file,
                                         white_label_id=white_label_id,
                                         primary_color=primary_color,
                                         logo_hidden=logo_hidden,
                                         team_name=team_name,
                                         workflow_name=workflow_name,
                                         stepready_email_text=stepready_email_text)
                                                     

        return dict_response

    def _file_endpoints(self, endpoint, file_name, get_request):
        token = self.access.get_jwt_token()
        request = self.request.get(endpoint, token)
        if get_request:
            return request
        return _file_downloader(file_name, request)

    def _post_workflow_instance(self, endpoint, **params):
        token = self.access.get_jwt_token()
        params = _clean_dict(params)

        response = self.request.post(endpoint, token, params)
        if response.status_code != 200:
            exc_msg = get_exception_message(response)
            raise Exception(exc_msg)
        return response.json().get('data')

    def _post_files(self, endpoint, **params):
        token = self.access.get_jwt_token()
        params = _clean_dict(params)
        response = self.request.post(endpoint, token, None, params)
        if response.status_code != 200:
            exc_msg = get_exception_message(response)
            raise Exception(exc_msg)
        return response.json().get('data')



#TODO
#Custom exceptions to throw when HW returns an error


def get_exception_message(response):
    response = response.json()
    if type(response) == str:
        exc_msg = response
    else:
        exc_msg = response.get('error', 'internal error')
    return exc_msg



import logging
import _io
logger = logging.getLogger(__name__)


def _file_downloader(file_name, file):
    with open(file_name, "wb") as f:
        try:
            for chunk in file.iter_content(chunk_size=8192): 
                if chunk:
                    f.write(chunk)
            else:
                return True
        except:
            logger.error('Could not write file')
        return False


def _clean_dict(dictionary):
    clean_dict = {**dictionary}
    for key, value in dictionary.items():
        if not value:
            del clean_dict[key]
    return clean_dict


def _file_extension(file, name):
    '''
    This method returns the extension of a file, or an exception if the object
    is not a file or if it has a wrong extension
    '''

    if type(file) not in [_io.BufferedReader, _io.BytesIO]:
        raise Exception("Wrong object type")

    extensions = ['png', 'jpeg','jpg', 'gif']
    extension = 'png'
    name = getattr(file, 'name', name)

    if not name:
        raise Exception("For this file object you need to include the logo_name property")


    extension = name.split('.')[-1]

    if extension not in extensions:
        raise Exception(f"Wrong file extension, must be one of {extensions}")
    return extension



from setuptools import setup


def readme():
    with open('README.md') as f:
        return f.read()


setup(
    name='helloworks-python-sdk',
    version='0.4',
    description="A Python wrapper for the HelloWorks API  (https://portal.helloworks.com/)",
    long_description=readme(),
    long_description_content_type="text/markdown",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8"
    ],
    keywords='Helloworks python api sdk',
    url='https://github.com/kevteg/helloworks-python-sdk',
    author='kevteg',
    author_email='kevteg05@gmail.com',
    license='MIT',
    packages=[
        'helloworks',
        'helloworks.utils',
    ],
    install_requires=[
        'requests'
    ],
    test_suite='nose.collector',
    tests_require=['nose'],
    include_package_data=True,
    zip_safe=False
)



import requests
import json


class HwRequest(object):

    def _process_request(self,
                         method,
                         url,
                         token,
                         data=None,
                         files=None):
        headers = {'Authorization': f'Bearer {token}'}
        request_method = getattr(requests, method)
        return request_method(url, json=data, headers=headers, files=files)

    def get(self, url, token, data=None):
        return self._process_request('get', url, token, data)

    def post(self, url, token, data=None, files=None):
        return self._process_request('post', url, token, data, files)

    def delete(self, url, token, data=None):
        return self._process_request('delete', url, token, data)

    def put(self, url, token, data=None):
        return self._process_request('put', url, token, data)



from .api_endpoints import ApiEndpoints
from .request import HwRequest
from .access import HwAccess
from .helper import _file_downloader, _clean_dict, _file_extension
__all__ = [ApiEndpoints, HwRequest, HwAccess]



from .request import HwRequest
from .api_endpoints import ApiEndpoints
from time import time


class HwAccess(object):
    def __init__(self, api_key_id, API_KEY_VALUE):
        self.request = HwRequest()
        self.endpoints = ApiEndpoints()
        self.api_key_id = api_key_id
        self.API_KEY_VALUE = API_KEY_VALUE
        self.expires_at = 0
        self.jwt_token = None

    def get_jwt_token(self):
        '''
        Method to get the latest JWT token
        '''
        now = int(time())
        if self.jwt_token and self.expires_at > now:
            return self.jwt_token
        else:
            self.expires_at, self.jwt_token = self._get_jwt_token()
        return self.jwt_token

    def _get_jwt_token(self):
        auth_token_endpoint = self.endpoints.auth_token(self.api_key_id)
        response = self.request.get(auth_token_endpoint, self.API_KEY_VALUE)
        return self._process_response(response)

    def _process_response(self, response):
        if response.status_code == 200:
            data = response.json().get('data', {})
            return data.get('expires_at'), data.get('token')
        else:
            raise Exception(response.json().get('error', 'Internal error'))
